import base64
from dataclasses import dataclass
from datetime import datetime
from functools import reduce
import logging
import math
import re
from time import time
from xml.etree import ElementTree
from flask import g
from argus.backend.models.web import ArgusEventTypes
from argus.backend.plugins.sct.testrun import SCTJunitReports, SCTTestRun, SubtestType
from argus.backend.plugins.sct.types import GeminiResultsRequest, PerformanceResultsRequest, ResourceUpdateRequest
from argus.backend.plugins.sct.udt import (
    CloudInstanceDetails,
    CloudResource,
    EventsBySeverity,
    NemesisRunInfo,
    NodeDescription,
    PackageVersion,
    PerformanceHDRHistogram,
)
from argus.backend.service.event_service import EventService
from argus.backend.util.common import get_build_number
from argus.backend.util.enums import NemesisStatus, ResourceState, TestStatus

LOGGER = logging.getLogger(__name__)


class SCTServiceException(Exception):
    pass


@dataclass(init=True, repr=True)
class NemesisSubmissionRequest:
    name: str
    class_name: str
    start_time: int
    node_name: str
    node_ip: str
    node_shards: int


@dataclass(init=True, repr=True)
class NemesisFinalizationRequest:
    name: str
    start_time: int
    status: str
    message: str


@dataclass(init=True, repr=True)
class EventSubmissionRequest:
    severity: str
    total_events: int
    messages: list[str]


class SCTService:

    @staticmethod
    def submit_packages(run_id: str, packages: list[dict]) -> str:
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            for package_dict in packages:
                package = PackageVersion(**package_dict)
                if package not in run.packages:
                    run.packages.append(package)
            run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "added"


    @staticmethod
    def set_sct_runner(run_id: str, public_ip: str, private_ip: str, region: str, backend: str):
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            run.sct_runner_host = CloudInstanceDetails(
                public_ip=public_ip,
                private_ip=private_ip,
                provider=backend,
                region=region,
            )
            run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "updated"

    @staticmethod
    def submit_screenshots(run_id: str, screenshot_links: list[str]) -> str:
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            for link in screenshot_links:
                run.add_screenshot(link)
            run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "submitted"

    @staticmethod
    def submit_gemini_results(run_id: str, gemini_data: GeminiResultsRequest) -> str:
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            run.subtest_name = SubtestType.GEMINI.value
            run.oracle_nodes_count = gemini_data.get("oracle_nodes_count")
            run.oracle_node_ami_id = gemini_data.get("oracle_node_ami_id")
            run.oracle_node_instance_type = gemini_data.get("oracle_node_instance_type")
            run.oracle_node_scylla_version = gemini_data.get("oracle_node_scylla_version")
            run.gemini_command = gemini_data.get("gemini_command")
            run.gemini_version = gemini_data.get("gemini_version")
            run.gemini_status = gemini_data.get("gemini_status")
            run.gemini_seed = str(gemini_data.get("gemini_seed"))
            run.gemini_write_ops = gemini_data.get("gemini_write_ops")
            run.gemini_write_errors = gemini_data.get("gemini_write_errors")
            run.gemini_read_ops = gemini_data.get("gemini_read_ops")
            run.gemini_read_errors = gemini_data.get("gemini_read_errors")
            run.save()

            if run.gemini_status != "PASSED":
                run.status = TestStatus.FAILED
                EventService.create_run_event(kind=ArgusEventTypes.TestRunStatusChanged, body={
                        "message": "[{username}] Setting run status to {status} due to Gemini reporting following status: {gemini_status}",
                        "username": g.user.username,
                        "status": TestStatus.FAILED.value,
                        "gemini_status": run.gemini_status,
                }, user_id=g.user.id, run_id=run_id, release_id=run.release_id, test_id=run.test_id)
                run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "submitted"

    @staticmethod
    def submit_performance_results(run_id: str, performance_results: PerformanceResultsRequest):
        # pylint: disable=too-many-statements
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            run.subtest_name = SubtestType.PERFORMANCE.value
            run.perf_op_rate_average = performance_results.get("perf_op_rate_average")
            run.perf_op_rate_total = performance_results.get("perf_op_rate_total")
            run.perf_avg_latency_99th = performance_results.get("perf_avg_latency_99th")
            run.perf_avg_latency_mean = performance_results.get("perf_avg_latency_mean")
            run.perf_total_errors = performance_results.get("perf_total_errors")
            run.stress_cmd = performance_results.get("stress_cmd")
            run.test_name = performance_results.get("test_name")
            run.save()

            is_latency_test = "latency" in run.test_name
            threshold_negative = -10

            def cmp(lhs, rhs):
                delta = rhs - lhs
                change = int(math.fabs(delta) * 100 / rhs)
                return change if delta >= 0 else change * -1

            previous_runs = SCTTestRun.get_perf_results_for_test_name(run.build_id, run.start_time, run.test_name)
            metrics_to_check = ["perf_avg_latency_99th", "perf_avg_latency_mean"] if is_latency_test else ["perf_op_rate_total"]

            older_runs_by_version = {}
            for prev_run in previous_runs:
                if not older_runs_by_version.get(prev_run["scylla_version"]):
                    older_runs_by_version[prev_run["scylla_version"]] = []
                older_runs_by_version[prev_run["scylla_version"]].append(prev_run)

            regression_found = False
            regression_info = {
                "version": None,
                "delta": None,
                "id": None,
                "metric": None,
                "job_url": None,
            }

            if performance_results["histograms"]:
                for histogram in performance_results["histograms"]:
                    run.histograms = { k: PerformanceHDRHistogram(**v) for k, v in histogram.items() }

            for version, runs in older_runs_by_version.items():
                for metric in metrics_to_check:
                    # pylint: disable=cell-var-from-loop
                    best_run = sorted(runs, reverse=(not is_latency_test), key=lambda v: v[metric])[0]
                    last_run = runs[0]

                    metric_to_best = cmp(run[metric], best_run[metric])
                    metric_to_last = cmp(run[metric], last_run[metric])
                    if metric_to_last < threshold_negative:
                        regression_found = True
                        regression_info["metric"] = metric
                        regression_info["version"] = version
                        regression_info["job_url"] = last_run["build_job_url"]
                        regression_info["id"] = str(last_run["id"])
                        regression_info["delta"] = metric_to_last
                        break

                    if metric_to_best < threshold_negative:
                        regression_found = True
                        regression_info["metric"] = metric
                        regression_info["version"] = version
                        regression_info["job_url"] = best_run["build_job_url"]
                        regression_info["id"] = str(best_run["id"])
                        regression_info["delta"] = metric_to_best
                        break

                if regression_found:
                    break

            if regression_found:
                run.status = TestStatus.FAILED.value
                run.save()
                EventService.create_run_event(kind=ArgusEventTypes.TestRunStatusChanged, body={
                        "message": "[{username}] Setting run status to {status} due to performance metric '{metric}' falling "
                                   "below allowed threshold ({threshold_negative}): {delta}% compared to "
                                   "<a href='/test/{test_id}/runs?additionalRuns[]={base_run_id}&additionalRuns[]={previous_run_id}'>This {version} (#{build_number}) run</a>",
                        "username": g.user.username,
                        "status": TestStatus.FAILED.value,
                        "metric": regression_info["metric"],
                        "threshold_negative": threshold_negative,
                        "delta": regression_info["delta"],
                        "test_id": str(run.test_id),
                        "base_run_id": str(run.id),
                        "previous_run_id": regression_info["id"],
                        "version": regression_info["version"],
                        "build_number": get_build_number(regression_info["job_url"])
                }, user_id=g.user.id, run_id=run_id, release_id=run.release_id, test_id=run.test_id)
            else:
                # NOTE: This will override status set by SCT Events.
                run.status = TestStatus.PASSED.value
                run.save()

        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "submitted"

    @staticmethod
    def get_performance_history_for_test(run_id: str):
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            rows = run.get_perf_results_for_test_name(build_id=run.build_id, start_time=run.start_time, test_name=run.test_name)
            return rows
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception


    @staticmethod
    def create_resource(run_id: str, resource_details: dict) -> str:
        instance_details = CloudInstanceDetails(**resource_details.pop("instance_details"))
        resource = CloudResource(**resource_details, instance_info=instance_details)
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            run.get_resources().append(resource)
            run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "created"

    @staticmethod
    def update_resource_shards(run_id: str, resource_name: str, new_shards: int) -> str:
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            resource = next(res for res in run.get_resources() if res.name == resource_name)
            resource.get_instance_info().shards_amount = new_shards
            run.save()
        except StopIteration as exception:
            LOGGER.error("Resource %s not found in run %s", resource_name, run_id)
            raise SCTServiceException("Resource not found", resource_name) from exception
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "updated"

    @staticmethod
    def update_resource(run_id: str, resource_name: str, update_data: ResourceUpdateRequest) -> str:
        try:
            fields_updated = {}
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            resource = next(res for res in run.get_resources() if res.name == resource_name)
            instance_info = update_data.pop("instance_info", None)
            resource.state = ResourceState(update_data.get("state", resource.state)).value
            if instance_info:
                resource_instance_info = resource.get_instance_info()
                for k, v in instance_info.items():
                    if k in resource_instance_info.keys():
                        resource_instance_info[k] = v
                        fields_updated[k] = v
            run.save()
        except StopIteration as exception:
            LOGGER.error("Resource %s not found in run %s", resource_name, run_id)
            raise SCTServiceException("Resource not found", resource_name) from exception
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return {
            "state": "updated",
            "fields": fields_updated
        }

    @staticmethod
    def terminate_resource(run_id: str, resource_name: str, reason: str) -> str:
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            resource = next(res for res in run.get_resources() if res.name == resource_name)
            resource.get_instance_info().termination_reason = reason
            resource.get_instance_info().termination_time = int(time())
            resource.state = ResourceState.TERMINATED.value
            run.save()
        except StopIteration as exception:
            LOGGER.error("Resource %s not found in run %s", resource_name, run_id)
            raise SCTServiceException("Resource not found", resource_name) from exception
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "terminated"

    @staticmethod
    def submit_nemesis(run_id: str, nemesis_details: dict) -> str:
        nem_req = NemesisSubmissionRequest(**nemesis_details)
        node_desc = NodeDescription(name=nem_req.node_name, ip=nem_req.node_ip, shards=nem_req.node_shards)
        nemesis_info = NemesisRunInfo(
            class_name=nem_req.class_name,
            name=nem_req.name,
            start_time=int(nem_req.start_time),
            end_time=0,
            duration=0,
            stack_trace="",
            status=NemesisStatus.RUNNING.value,
            target_node=node_desc,
        )
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            run.add_nemesis(nemesis_info)
            run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "created"

    @staticmethod
    def finalize_nemesis(run_id: str, nemesis_details: dict) -> str:
        nem_req = NemesisFinalizationRequest(**nemesis_details)
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            nemesis = next(nem for nem in run.get_nemeses() if nem.name ==
                           nem_req.name and nem.start_time == nem_req.start_time)
            nemesis.status = NemesisStatus(nem_req.status).value
            nemesis.stack_trace = nem_req.message
            nemesis.end_time = int(time())
            run.save()
        except StopIteration as exception:
            LOGGER.error("Nemesis %s (%s) not found for run %s", nem_req.name, nem_req.start_time, run_id)
            raise SCTServiceException("Nemesis not found", (nem_req.name, nem_req.start_time)) from exception
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "updated"

    @staticmethod
    def submit_events(run_id: str, events: list[dict]) -> str:
        wrapped_events = [EventSubmissionRequest(**ev) for ev in events]
        try:
            run: SCTTestRun = SCTTestRun.get(id=run_id)
            for event in wrapped_events:
                wrapper = EventsBySeverity(severity=event.severity,
                                           event_amount=event.total_events, last_events=event.messages)
                run.get_events().append(wrapper)
            coredumps = SCTService.locate_coredumps(run, run.get_events())
            run.submit_logs(coredumps)
            run.save()
        except SCTTestRun.DoesNotExist as exception:
            LOGGER.error("Run %s not found for SCTTestRun", run_id)
            raise SCTServiceException("Run not found", run_id) from exception

        return "added"

    @staticmethod
    def locate_coredumps(run: SCTTestRun, events: list[EventsBySeverity]) -> list[dict]:
        flat_messages: list[str] = []
        links = []
        for es in events:
            flat_messages.extend(es.last_events)
        coredump_events = filter(lambda v: "coredumpevent" in v.lower(), flat_messages)
        for idx, event in enumerate(coredump_events):
            core_pattern = r"corefile_url=(?P<url>.+)$"
            ts_pattern = r"^(?P<ts>\d{4}-\d{2}-\d{2} ([\d:]*)\.\d{3})"
            node_name_pattern = r"node=(?P<name>.+)$"
            core_url_match = re.search(core_pattern, event, re.MULTILINE)
            node_name_match = re.search(node_name_pattern, event, re.MULTILINE)
            ts_match = re.search(ts_pattern, event)
            if core_url_match:
                node_name = node_name_match.group("name") if node_name_match else f"unknown-node-{idx}"
                split_name = node_name.split(" ")
                node_name = split_name[1] if len(split_name) >= 2 else node_name
                url = core_url_match.group("url")
                timestamp_component = ""
                if ts_match:
                    try:
                        timestamp = datetime.fromisoformat(ts_match.group("ts"))
                        timestamp_component = timestamp.strftime("-%Y-%m-%d_%H-%M-%S")
                    except ValueError:
                        pass
                log_link = {
                    "log_name": f"core.scylla-{node_name}{timestamp_component}.gz",
                    "log_link": url
                }
                links.append(log_link)
        return links

    @staticmethod
    def get_scylla_version_kernels_report(release_name: str):
        all_release_runs = SCTTestRun.get_version_data_for_release(release_name=release_name)
        kernels_by_version = {}
        kernel_metadata = {}
        for run in all_release_runs:
            packages = run["packages"]
            if not packages:
                continue
            scylla_pkgs = {p["name"]: p for p in packages if "scylla-server" in p["name"]}
            scylla_pkg = scylla_pkgs["scylla-server-upgraded"] if scylla_pkgs.get(
                "scylla-server-upgraded") else scylla_pkgs.get("scylla-server")
            version = f"{scylla_pkg['version']}-{scylla_pkg['date']}.{scylla_pkg['revision_id']}" if scylla_pkgs else "unknown"
            kernel_packages = [p for p in packages if "kernel" in p["name"]]
            kernel_package = kernel_packages[0] if len(kernel_packages) > 0 else None
            if not kernel_package:
                continue
            version_list = set(kernels_by_version.get(version, []))
            version_list.add(kernel_package["version"])
            kernels_by_version[version] = list(version_list)
            metadata = kernel_metadata.get(
                kernel_package.version,
                {
                    "passed": 0,
                    "failed": 0,
                    "aborted": 0,
                }
            )
            if run["status"] in ["passed", "failed", "aborted", "test_error"]:
                metadata[run["status"]] += 1
            kernel_metadata[kernel_package["version"]] = metadata

        return {
            "versions": kernels_by_version,
            "metadata": kernel_metadata
        }
    
    @staticmethod
    def junit_submit(run_id: str, file_name: str, content: str) -> bool:
        try:
            report = SCTJunitReports.get(test_id=run_id, file_name=file_name)
            if report:
                raise SCTServiceException(f"Report {file_name} already exists.", file_name)
        except SCTJunitReports.DoesNotExist:
            pass
        report = SCTJunitReports()
        report.test_id = run_id
        report.file_name = file_name
        
        xml_content = str(base64.decodebytes(bytes(content, encoding="utf-8")), encoding="utf-8")
        try:
            _ = ElementTree.fromstring(xml_content)
        except Exception:
            raise SCTServiceException(f"Malformed JUnit report submitted")

        report.report = xml_content
        report.save()

        return True

    @staticmethod
    def junit_get_all(run_id: str) -> list[SCTJunitReports]:
        return list(SCTJunitReports.filter(test_id=run_id).all())
    
    @staticmethod
    def junit_get_single(run_id: str, file_name: str) -> SCTJunitReports:
        return SCTJunitReports.get(test_id=run_id, file_name=file_name)


from uuid import UUID
from argus.backend.db import ScyllaCluster
from argus.backend.models.result import ArgusGenericResultMetadata, ArgusGenericResultData
from argus.backend.plugins.core import PluginModelBase
from argus.backend.plugins.loader import AVAILABLE_PLUGINS
from argus.backend.util.enums import TestStatus


class ClientException(Exception):
    pass


class ClientService:
    PLUGINS = {name: plugin.model for name, plugin in AVAILABLE_PLUGINS.items()}

    def __init__(self) -> None:
        self.cluster = ScyllaCluster.get()

    def get_model(self, run_type: str) -> PluginModelBase:
        cls = self.PLUGINS.get(run_type)
        if not cls:
            raise ClientException(f"Unsupported run type: {run_type}", run_type)
        return cls

    def submit_run(self, run_type: str, request_data: dict) -> str:
        model = self.get_model(run_type)
        model.submit_run(request_data=request_data)
        return "Created"

    def heartbeat(self, run_type: str, run_id: str) -> int:
        model = self.get_model(run_type)
        run = model.load_test_run(UUID(run_id))
        run.update_heartbeat()
        run.save()
        return run.heartbeat

    def get_run_status(self, run_type: str, run_id: str) -> str:
        model = self.get_model(run_type)
        run = model.load_test_run(UUID(run_id))
        return run.status

    def update_run_status(self, run_type: str, run_id: str, new_status: str) -> str:
        model = self.get_model(run_type)
        run = model.load_test_run(UUID(run_id))
        run.change_status(new_status=TestStatus(new_status))
        run.save()

        return run.status

    def submit_product_version(self, run_type: str, run_id: str, version: str) -> str:
        model = self.get_model(run_type)
        run = model.load_test_run(UUID(run_id))
        run.submit_product_version(version)
        run.save()

        return "Submitted"

    def submit_logs(self, run_type: str, run_id: str, logs: list[dict]) -> str:
        model = self.get_model(run_type)
        run = model.load_test_run(UUID(run_id))
        run.submit_logs(logs)
        run.save()

        return "Submitted"

    def finish_run(self, run_type: str, run_id: str, payload: dict | None = None) -> str:
        model = self.get_model(run_type)
        run = model.load_test_run(UUID(run_id))
        run.finish_run(payload)
        run.save()

        return "Finalized"

    def submit_results(self, run_type: str, run_id: str, results: dict) -> str:
        model = self.get_model(run_type)
        run = model.load_test_run(UUID(run_id))
        ArgusGenericResultMetadata(test_id=run.test_id, **results["meta"]).save()
        if results.get("sut_timestamp", 0) == 0:
            results["sut_timestamp"] = run.sut_timestamp()  # automatic sut_timestamp
        table_name = results["meta"]["name"]
        sut_timestamp = results["sut_timestamp"]
        for cell in results["results"]:
            ArgusGenericResultData(test_id=run.test_id,
                                   run_id=run.id,
                                   name=table_name,
                                   sut_timestamp=sut_timestamp,
                                   **cell
                                   ).save()
        return "Submitted"



from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Union


class Status(Enum):
    PASS = auto()
    WARNING = auto()
    ERROR = auto()

    def __str__(self):
        return self.name


class ResultType(Enum):
    INTEGER = auto()
    FLOAT = auto()
    DURATION = auto()

    def __str__(self):
        return self.name


@dataclass
class ColumnMetadata:
    name: str
    unit: str
    type: ResultType

    def as_dict(self) -> dict:
        return {
            "name": self.name,
            "unit": self.unit,
            "type": str(self.type)
        }


class ResultTableMeta(type):
    def __new__(cls, name, bases, dct):
        cls_instance = super().__new__(cls, name, bases, dct)
        meta = dct.get('Meta')

        if meta:
            cls_instance.name = meta.name
            cls_instance.description = meta.description
            cls_instance.columns = meta.Columns
            cls_instance.column_names = {column.name for column in cls_instance.columns}
            cls_instance.rows = []
        return cls_instance


@dataclass
class Cell:
    column: str
    row: str
    value: Union[int, float, str]
    status: Status

    def as_dict(self) -> dict:
        return {
            "column": self.column,
            "row": self.row,
            "value": self.value,
            "status": str(self.status)
        }


@dataclass
class GenericResultTable(metaclass=ResultTableMeta):
    """
    Base class for all Generic Result Tables in Argus. Use it as a base class for your result table.
    """
    sut_timestamp: int = 0  # automatic timestamp based on SUT version. Works only with SCT and refers to Scylla version.
    sut_details: str = ""
    results: list[Cell] = field(default_factory=list)

    def as_dict(self) -> dict:
        rows = []
        for result in self.results:
            if result.row not in rows:
                rows.append(result.row)

        meta_info = {
            "name": self.name,
            "description": self.description,
            "columns_meta": [column.as_dict() for column in self.columns],
            "rows_meta": rows
        }
        return {
            "meta": meta_info,
            "sut_timestamp": self.sut_timestamp,
            "sut_details": self.sut_details,
            "results": [result.as_dict() for result in self.results]
        }

    def add_result(self, column: str, row: str, value: Union[int, float, str], status: Status):
        if column not in self.column_names:
            raise ValueError(f"Column {column} not found in the table")
        self.results.append(Cell(column=column, row=row, value=value, status=status))



from collections import defaultdict
from datetime import datetime, timedelta
from functools import reduce
import json
import logging
import re
import time
from typing import Any
from uuid import UUID

import requests
from flask import g
from cassandra.query import BatchStatement, ConsistencyLevel
from cassandra.cqlengine.query import BatchQuery
from argus.backend.db import ScyllaCluster
from argus.backend.models.result import ArgusGenericResultMetadata, ArgusGenericResultData

from argus.backend.models.web import (
    ArgusEvent,
    ArgusEventTypes,
    ArgusGithubIssue,
    ArgusGroup,
    ArgusNotificationSourceTypes,
    ArgusNotificationTypes,
    ArgusRelease,
    ArgusTest,
    ArgusTestRunComment,
    ArgusUserView,
    User,
    UserOauthToken,
)

from argus.backend.plugins.core import PluginInfoBase, PluginModelBase

from argus.backend.plugins.loader import AVAILABLE_PLUGINS
from argus.backend.events.event_processors import EVENT_PROCESSORS
from argus.backend.service.event_service import EventService
from argus.backend.service.notification_manager import NotificationManagerService
from argus.backend.service.stats import ComparableTestStatus
from argus.backend.util.common import chunk, get_build_number, strip_html_tags
from argus.backend.util.enums import TestInvestigationStatus, TestStatus

LOGGER = logging.getLogger(__name__)


class TestRunServiceException(Exception):
    pass


class TestRunService:
    ASSIGNEE_PLACEHOLDER = "none-none-none"

    RE_MENTION = r"@[A-Za-z\d](?:[A-Za-z\d]|-(?=[A-Za-z\d])){0,38}"

    plugins = AVAILABLE_PLUGINS
    github_headers = {
        "Accept": "application/vnd.github.v3+json",
    }

    def __init__(self) -> None:
        self.notification_manager = NotificationManagerService()

    def get_plugin(self, plugin_name: str) -> PluginInfoBase | None:
        return self.plugins.get(plugin_name)

    def get_run(self, run_type: str, run_id: UUID) -> PluginModelBase:
        plugin = self.plugins.get(run_type)
        if plugin:
            try:
                return plugin.model.get(id=run_id)
            except plugin.model.DoesNotExist:
                return None

    def get_runs_by_test_id(self, test_id: UUID, additional_runs: list[UUID], limit: int = 10):
        test: ArgusTest = ArgusTest.get(id=test_id)
        plugin = self.get_plugin(plugin_name=test.plugin_name)
        if not plugin:
            return []

        last_runs: list[dict] = plugin.model.get_run_meta_by_build_id(build_id=test.build_system_id, limit=limit)
        last_runs_ids = [run["id"] for run in last_runs]
        for added_run in additional_runs:
            if added_run not in last_runs_ids:
                last_runs.extend(plugin.model.get_run_meta_by_run_id(run_id=added_run))

        for row in last_runs:
            row["build_number"] = get_build_number(build_job_url=row["build_job_url"])

        last_runs = sorted(last_runs, reverse=True, key=lambda run: (run["build_number"], ComparableTestStatus(TestStatus(run["status"]))))

        return last_runs

    def get_runs_by_id(self, test_id: UUID, runs: list[UUID]):  # FIXME: Not needed, use get_run and individual polling
        # This is a batch request.
        test = ArgusTest.get(id=test_id)
        plugin = self.get_plugin(plugin_name=test.plugin_name)
        polled_runs: list[PluginModelBase] = []
        for run_id in runs:
            try:
                run: PluginModelBase = plugin.model.get(id=run_id)
                polled_runs.append(run)
            except plugin.model.DoesNotExist:
                pass

        response = {str(run.id): run for run in polled_runs}
        return response

    def change_run_status(self, test_id: UUID, run_id: UUID, new_status: TestStatus):
        try:
            test = ArgusTest.get(id=test_id)
        except ArgusTest.DoesNotExist as exc:
            raise TestRunServiceException("Test entity does not exist for provided test_id", test_id) from exc
        plugin = self.get_plugin(plugin_name=test.plugin_name)
        run: PluginModelBase = plugin.model.get(id=run_id)
        old_status = run.status
        run.status = new_status.value
        run.save()

        EventService.create_run_event(
            kind=ArgusEventTypes.TestRunStatusChanged,
            body={
                "message": "Status was changed from {old_status} to {new_status} by {username}",
                "old_status": old_status,
                "new_status": new_status.value,
                "username": g.user.username
            },
            user_id=g.user.id,
            run_id=run.id,
            release_id=test.release_id,
            group_id=test.group_id,
            test_id=test.id
        )

        return {
            "test_run_id": run.id,
            "status": new_status
        }

    def change_run_investigation_status(self, test_id: UUID, run_id: UUID, new_status: TestInvestigationStatus):
        test = ArgusTest.get(id=test_id)
        plugin = self.get_plugin(plugin_name=test.plugin_name)
        run: PluginModelBase = plugin.model.get(id=run_id)
        old_status = run.investigation_status
        run.investigation_status = new_status.value
        run.save()

        EventService.create_run_event(
            kind=ArgusEventTypes.TestRunStatusChanged,
            body={
                "message": "Investigation status was changed from {old_status} to {new_status} by {username}",
                "old_status": old_status,
                "new_status": new_status.value,
                "username": g.user.username
            },
            user_id=g.user.id,
            run_id=run.id,
            release_id=test.release_id,
            group_id=test.group_id,
            test_id=test.id
        )

        return {
            "test_run_id": run.id,
            "investigation_status": new_status
        }

    def change_run_assignee(self, test_id: UUID, run_id: UUID, new_assignee: UUID | None):
        test = ArgusTest.get(id=test_id)
        plugin = self.get_plugin(plugin_name=test.plugin_name)
        if not plugin:
            return {
                "test_run_id": run.id,
                "assignee": None
            }

        run: PluginModelBase = plugin.model.get(id=run_id)
        old_assignee = run.assignee
        run.assignee = new_assignee
        run.save()

        if new_assignee:
            new_assignee_user = User.get(id=new_assignee)
        else:
            new_assignee_user = None
        if old_assignee:
            try:
                old_assignee_user = User.get(id=old_assignee)
            except User.DoesNotExist:
                LOGGER.warning("Non existent assignee was present on the run %s for test %s: %s",
                               run_id, test_id, old_assignee)
                old_assignee = None
        EventService.create_run_event(
            kind=ArgusEventTypes.AssigneeChanged,
            body={
                "message": "Assignee was changed from \"{old_user}\" to \"{new_user}\" by {username}",
                "old_user": old_assignee_user.username if old_assignee else "None",
                "new_user": new_assignee_user.username if new_assignee else "None",
                "username": g.user.username
            },
            user_id=g.user.id,
            run_id=run.id,
            release_id=test.release_id,
            group_id=test.group_id,
            test_id=test.id
        )
        return {
            "test_run_id": run.id,
            "assignee": str(new_assignee_user.id) if new_assignee_user else None
        }

    def get_run_comment(self, comment_id: UUID):
        try:
            return ArgusTestRunComment.get(id=comment_id)
        except ArgusTestRunComment.DoesNotExist:
            return None

    def get_run_comments(self, run_id: UUID):
        return sorted(ArgusTestRunComment.filter(test_run_id=run_id).all(), key=lambda c: c.posted_at)

    def post_run_comment(self, test_id: UUID, run_id: UUID, message: str, reactions: dict, mentions: list[str]):
        message_stripped = strip_html_tags(message)

        mentions = set(mentions)
        for potential_mention in re.findall(self.RE_MENTION, message_stripped):
            if user := User.exists_by_name(potential_mention.lstrip("@")):
                mentions.add(user)

        test: ArgusTest = ArgusTest.get(id=test_id)
        plugin = self.get_plugin(test.plugin_name)
        release: ArgusRelease = ArgusRelease.get(id=test.release_id)
        comment = ArgusTestRunComment()
        comment.test_id = test.id
        comment.message = message_stripped
        comment.reactions = reactions
        comment.mentions = [m.id for m in mentions]
        comment.test_run_id = run_id
        comment.release_id = release.id
        comment.user_id = g.user.id
        comment.posted_at = time.time()
        comment.save()

        run: PluginModelBase = plugin.model.get(id=run_id)
        build_number = get_build_number(build_job_url=run.build_job_url)
        for mention in mentions:
            params = {
                "username": g.user.username,
                "run_id": comment.test_run_id,
                "test_id": test.id,
                "build_id": run.build_id,
                "build_number": build_number,
            }
            self.notification_manager.send_notification(
                receiver=mention.id,
                sender=comment.user_id,
                notification_type=ArgusNotificationTypes.Mention,
                source_type=ArgusNotificationSourceTypes.Comment,
                source_id=comment.id,
                source_message=comment.message,
                content_params=params
            )

        EventService.create_run_event(kind=ArgusEventTypes.TestRunCommentPosted, body={
            "message": "A comment was posted by {username}",
            "username": g.user.username
        }, user_id=g.user.id, run_id=run_id, release_id=release.id, test_id=test.id)

        return self.get_run_comments(run_id=run_id)

    def delete_run_comment(self, comment_id: UUID, test_id: UUID, run_id: UUID):
        comment: ArgusTestRunComment = ArgusTestRunComment.get(id=comment_id)
        if comment.user_id != g.user.id:
            raise Exception("Unable to delete other user comments")
        comment.delete()

        EventService.create_run_event(kind=ArgusEventTypes.TestRunCommentDeleted, body={
            "message": "A comment was deleted by {username}",
            "username": g.user.username
        }, user_id=g.user.id, run_id=run_id, release_id=comment.release_id, test_id=test_id)

        return self.get_run_comments(run_id=run_id)

    def update_run_comment(self, comment_id: UUID, test_id: UUID, run_id: UUID, message: str, mentions: list[str], reactions: dict):
        comment: ArgusTestRunComment = ArgusTestRunComment.get(id=comment_id)
        if comment.user_id != g.user.id:
            raise Exception("Unable to edit other user comments")
        comment.message = strip_html_tags(message)
        comment.reactions = reactions
        comment.mentions = mentions
        comment.save()

        EventService.create_run_event(kind=ArgusEventTypes.TestRunCommentUpdated, body={
            "message": "A comment was edited by {username}",
            "username": g.user.username
        }, user_id=g.user.id, run_id=run_id, release_id=comment.release_id, test_id=test_id)

        return self.get_run_comments(run_id=run_id)

    def get_run_events(self, run_id: UUID):
        response = {}
        all_events = ArgusEvent.filter(run_id=run_id).all()
        all_events = sorted(all_events, key=lambda ev: ev.created_at)
        response["run_id"] = run_id
        response["raw_events"] = [dict(event.items()) for event in all_events]
        response["events"] = {
            str(event.id): EVENT_PROCESSORS.get(event.kind)(json.loads(event.body))
            for event in all_events
        }
        return response

    def fetch_results(self, test_id: UUID, run_id: UUID) -> dict:
        query_fields = ["column", "row", "value", "status"]
        tables_meta = ArgusGenericResultMetadata.filter(test_id=test_id)
        tables = []
        for table in tables_meta:
            cells = ArgusGenericResultData.objects.filter(test_id=test_id, run_id=run_id, name=table.name).only(query_fields)
            if not cells:
                continue
            tables.append({'meta': {
                'name': table.name,
                'description': table.description,
                'columns_meta': table.columns_meta,
                'rows_meta': table.rows_meta
            },
                'cells': [{k:v for k,v in cell.items() if k in query_fields} for cell in cells]})

        return tables

    def submit_github_issue(self, issue_url: str, test_id: UUID, run_id: UUID):
        user_tokens = UserOauthToken.filter(user_id=g.user.id).all()
        token = None
        for tok in user_tokens:
            if tok.kind == "github":
                token = tok.token
                break
        if not token:
            raise Exception("Github token not found")

        match = re.match(
            r"http(s)?://(www\.)?github\.com/(?P<owner>[\w\d]+)/"
            r"(?P<repo>[\w\d\-_]+)/(?P<type>issues|pull)/(?P<issue_number>\d+)(/)?",
            issue_url,
        )
        if not match:
            raise Exception("URL doesn't match Github schema")

        test: ArgusTest = ArgusTest.get(id=test_id)
        plugin = self.get_plugin(plugin_name=test.plugin_name)

        run = plugin.model.get(id=run_id)
        release = ArgusRelease.get(id=run["release_id"])
        test = ArgusTest.get(build_system_id=run["build_id"])
        group = ArgusGroup.get(id=test.group_id)

        new_issue = ArgusGithubIssue()
        new_issue.user_id = g.user.id
        new_issue.run_id = run_id
        new_issue.group_id = group.id
        new_issue.release_id = release.id
        new_issue.test_id = test.id
        new_issue.type = match.group("type")
        new_issue.owner = match.group("owner")
        new_issue.repo = match.group("repo")
        new_issue.issue_number = int(match.group("issue_number"))

        issue_request = requests.get(
            f"https://api.github.com/repos/{new_issue.owner}/{new_issue.repo}/issues/{new_issue.issue_number}",
            headers={
                **self.github_headers,
                "Authorization": f"token {token}",
            }
        )
        if issue_request.status_code != 200:
            raise Exception(f"Error getting issue state: Response: HTTP {issue_request.status_code}", issue_request.json())

        issue_state: dict[str, Any] = issue_request.json()

        new_issue.title = issue_state.get("title")
        new_issue.url = issue_state.get("html_url")
        new_issue.last_status = issue_state.get("state")
        new_issue.save()

        EventService.create_run_event(
            kind=ArgusEventTypes.TestRunIssueAdded,
            body={
                "message": "An issue titled \"{title}\" was added by {username}",
                "username": g.user.username,
                "url": issue_url,
                "title": issue_state.get("title"),
                "state": issue_state.get("state"),
            },
            user_id=g.user.id,
            run_id=new_issue.run_id,
            release_id=new_issue.release_id,
            group_id=new_issue.group_id,
            test_id=new_issue.test_id
        )

        response = {
            **dict(list(new_issue.items())),
            "title": issue_state.get("title"),
            "state": issue_state.get("state"),
        }

        return response

    def _get_github_issues_for_view(self, view_id: UUID | str) -> list[ArgusGithubIssue]:
        view: ArgusUserView = ArgusUserView.get(id=view_id)
        issues = []
        for batch in chunk(view.tests):
            issues.extend(ArgusGithubIssue.filter(test_id__in=batch).allow_filtering().all())
        
        return issues

    def get_github_issues(self, filter_key: str, filter_id: UUID, aggregate_by_issue: bool = False) -> dict:
        if filter_key not in ["release_id", "group_id", "test_id", "run_id", "user_id", "view_id"]:
            raise Exception(
                "filter_key can only be one of: \"release_id\", \"group_id\", \"test_id\", \"run_id\", \"user_id\", \"view_id\""
            )
        if filter_key == "view_id":
            all_issues = self._get_github_issues_for_view(filter_id)
        else:
            all_issues = ArgusGithubIssue.filter(**{filter_key: filter_id}).all()
        if aggregate_by_issue:
            runs_by_issue = {}
            response = []
            for issue in all_issues:
                runs = runs_by_issue.get(issue, [])
                runs.append({"test_id": issue.test_id, "run_id": issue.run_id})
                runs_by_issue[issue] = runs

            for issue, runs in runs_by_issue.items():
                issue_dict = dict(issue.items())
                issue_dict["runs"] = runs
                response.append(issue_dict)

        else:
            response = [dict(issue.items()) for issue in all_issues]
        return response

    def resolve_run_build_id_and_number_multiple(self, runs: list[tuple[UUID, UUID]]) -> dict[UUID, dict[str, Any]]:
        test_ids = [r[0] for r in runs]
        all_tests: list = []
        for id_slice in chunk(test_ids):
            all_tests.extend(ArgusTest.filter(id__in=id_slice).all())

        tests: dict[str, ArgusTest] = {str(t.id): t for t in all_tests}
        runs_by_plugin = reduce(lambda acc, val: acc[tests[val[0]].plugin_name].append(val[1]) or acc, runs, defaultdict(list))
        all_runs = {}
        for plugin, run_ids in runs_by_plugin.items():
            model = AVAILABLE_PLUGINS.get(plugin).model
            model_runs = []
            for run_id in run_ids:
                model_runs.append(model.filter(id=run_id).only(["build_id", "start_time", "build_job_url", "id", "test_id"]).get())
            all_runs.update({ str(run["id"]): {**run, "build_number": get_build_number(run["build_job_url"])} for run in model_runs })

        return all_runs


    def delete_github_issue(self, issue_id: UUID) -> dict:
        issue: ArgusGithubIssue = ArgusGithubIssue.get(id=issue_id)

        EventService.create_run_event(
            kind=ArgusEventTypes.TestRunIssueRemoved,
            body={
                "message": "An issue titled \"{title}\" was removed by {username}",
                "username": g.user.username,
                "url": issue.url,
                "title": issue.title,
                "state": issue.last_status,
            },
            user_id=g.user.id,
            run_id=issue.run_id,
            release_id=issue.release_id,
            group_id=issue.group_id,
            test_id=issue.test_id
        )
        issue.delete()

        return {
            "deleted": issue_id
        }

    def terminate_stuck_runs(self):
        sct = AVAILABLE_PLUGINS.get("scylla-cluster-tests").model
        now = datetime.utcnow()
        stuck_period = now - timedelta(minutes=45)
        stuck_runs_running = sct.filter(heartbeat__lt=int(
            stuck_period.timestamp()), status=TestStatus.RUNNING.value).allow_filtering().all()
        stuck_runs_created = sct.filter(heartbeat__lt=int(
            stuck_period.timestamp()), status=TestStatus.CREATED.value).allow_filtering().all()

        all_stuck_runs = [*stuck_runs_running, *stuck_runs_created]
        LOGGER.info("Found %s stuck runs", len(all_stuck_runs))

        for run in all_stuck_runs:
            LOGGER.info("Will set %s as ABORTED", run.id)
            old_status = run.status
            run.status = TestStatus.ABORTED.value
            run.save()

            EventService.create_run_event(
                kind=ArgusEventTypes.TestRunStatusChanged,
                body={
                    "message": "Run was automatically terminated due to not responding for more than 45 minutes "
                               "(Status changed from {old_status} to {new_status}) by {username}",
                    "old_status": old_status,
                    "new_status": run.status,
                    "username": g.user.username
                },
                user_id=g.user.id,
                run_id=run.id,
                release_id=run.release_id,
                group_id=run.group_id,
                test_id=run.test_id
            )

        return len(all_stuck_runs)

    def ignore_jobs(self, test_id: UUID, reason: str):
        test: ArgusTest = ArgusTest.get(id=test_id)
        plugin = self.get_plugin(plugin_name=test.plugin_name)

        if not reason:
            raise TestRunServiceException("Reason for ignore cannot be empty")

        cluster = ScyllaCluster.get()
        batch = BatchStatement(consistency_level=ConsistencyLevel.QUORUM)
        event_batch = BatchQuery()
        jobs_affected = 0
        for job in plugin.model.get_jobs_meta_by_test_id(test.id):
            if job["status"] != TestStatus.PASSED and job["investigation_status"] == TestInvestigationStatus.NOT_INVESTIGATED:
                batch.add(
                    plugin.model.prepare_investigation_status_update_query(
                        build_id=job["build_id"],
                        start_time=job["start_time"],
                        new_status=TestInvestigationStatus.IGNORED
                    )
                )

                ArgusEvent.batch(event_batch).create(
                    release_id = job["release_id"],
                    group_id = job["group_id"],
                    test_id = test_id,
                    user_id = g.user.id,
                    run_id = job["id"],
                    body = json.dumps({
                        "message": "Run was marked as ignored by {username} due to the following reason: {reason}",
                        "username": g.user.username,
                        "reason": reason,
                    }, ensure_ascii=True, separators=(',', ':')),
                    kind = ArgusEventTypes.TestRunBatchInvestigationStatusChange.value,
                    created_at = datetime.utcnow(),
                )

                jobs_affected += 1

        cluster.session.execute(batch)
        event_batch.execute()

        return jobs_affected



import logging
from uuid import UUID
from flask import (
    Blueprint,
    jsonify,
    request,
)
from argus.backend.error_handlers import handle_api_exception
from argus.backend.models.web import User
from argus.backend.service.stats import ViewStatsCollector
from argus.backend.service.user import api_login_required
from argus.backend.service.views import UserViewService
from argus.backend.util.common import get_payload

bp = Blueprint('view_api', __name__, url_prefix='/views')
LOGGER = logging.getLogger(__name__)
bp.register_error_handler(Exception, handle_api_exception)


class ViewApiException(Exception):
    pass


@bp.route("/", methods=["GET"])
@api_login_required
def index():
    return {
        "status": "ok",
        "response": {
            "version": "v1",
        }
    }


@bp.route("/create", methods=["POST"])
@api_login_required
def create_view():
    payload = get_payload(request)
    service = UserViewService()
    view = service.create_view(
        name=payload["name"], 
        items=payload["items"],
        widget_settings=payload["settings"],
        description=payload.get("description"),
        display_name=payload.get("displayName")
    )
    return {
        "status": "ok",
        "response": view
    }


@bp.route("/get", methods=["GET"])
@api_login_required
def get_view():
    view_id = request.args.get("viewId")
    if not view_id:
        raise ViewApiException("No viewId provided.")
    service = UserViewService()
    view = service.get_view(UUID(view_id))
    return {
        "status": "ok",
        "response": view
    }


@bp.route("/all", methods=["GET"])
@api_login_required
def get_all_views():
    user_id = request.args.get("userId")
    if user_id:
        user = User.get(id=user_id)
    else:
        user = None
    service = UserViewService()
    views = service.get_all_views(user)
    return {
        "status": "ok",
        "response": views
    }


@bp.route("/update", methods=["POST"])
@api_login_required
def update_view():
    payload = get_payload(request)
    service = UserViewService()
    res = service.update_view(view_id=payload["viewId"], update_data=payload["updateData"])
    return {
        "status": "ok",
        "response": res
    }


@bp.route("/delete", methods=["POST"])
@api_login_required
def delete_view():
    payload = get_payload(request)
    service = UserViewService()
    res = service.delete_view(payload["viewId"])
    return {
        "status": "ok",
        "response": res
    }


@bp.route("/search", methods=["GET"])
@api_login_required
def search_tests():
    query = request.args.get("query")
    service = UserViewService()
    if query:
        res = service.test_lookup(query)
    else:
        res = []
    return {
        "status": "ok",
        "response": {
            "hits": res,
            "total": len(res)
        }
    }

@bp.route("/stats", methods=["GET"])
@api_login_required
def view_stats():
    view_id = request.args.get("viewId")
    if not view_id:
        raise ViewApiException("No view id provided.")
    limited = bool(int(request.args.get("limited", 0)))
    version = request.args.get("productVersion", None)
    include_no_version = bool(int(request.args.get("includeNoVersion", True)))
    force = bool(int(request.args.get("force",  0)))
    collector = ViewStatsCollector(view_id=view_id, filter=version)
    stats = collector.collect(limited=limited, force=force, include_no_version=include_no_version)

    res = jsonify({
        "status": "ok",
        "response": stats
    })
    res.cache_control.max_age = 300
    return res

@bp.route("/<string:view_id>/versions", methods=["GET"])
@api_login_required
def view_versions(view_id: str):
    service = UserViewService()
    res = service.get_versions_for_view(view_id)
    return {
        "status": "ok",
        "response": res
    }

@bp.route("/<string:view_id>/resolve", methods=["GET"])
@api_login_required
def view_resolve(view_id: str):
    service = UserViewService()
    res = service.resolve_view_for_edit(view_id)
    return {
        "status": "ok",
        "response": res
    }


import re
import requests
from typing import Any, TypedDict
from uuid import UUID
import xml.etree.ElementTree as ET
import jenkins
import logging

from flask import current_app, g

from argus.backend.models.web import ArgusGroup, ArgusRelease, ArgusTest, UserOauthToken

LOGGER = logging.getLogger(__name__)
GITHUB_REPO_RE = r"(?P<http>^https?:\/\/(www\.)?github\.com\/(?P<user>[\w\d\-]+)\/(?P<repo>[\w\d\-]+)(\.git)?$)|(?P<ssh>git@github\.com:(?P<ssh_user>[\w\d\-]+)\/(?P<ssh_repo>[\w\d\-]+)(\.git)?)"

class Parameter(TypedDict):
    _class: str
    name: str
    description: str
    value: Any


class JenkinsServiceError(Exception):
    pass


class JenkinsService:
    RESERVED_PARAMETER_NAME = "requested_by_user"

    SETTINGS_CONFIG_MAP = {
        "scylla-cluster-tests": {
            "gitRepo": "*//scm/userRemoteConfigs/hudson.plugins.git.UserRemoteConfig/url",
            "gitBranch": "*//scm/branches/hudson.plugins.git.BranchSpec/name",
            "pipelineFile": "*//scriptPath",
        },
        "driver-matrix-tests": {
            "gitRepo": "*//scm/userRemoteConfigs/hudson.plugins.git.UserRemoteConfig/url",
            "gitBranch": "*//scm/branches/hudson.plugins.git.BranchSpec/name",
            "pipelineFile": "*//scriptPath",
        },
        "sirenada": {
            "gitRepo": "*//scm/userRemoteConfigs/hudson.plugins.git.UserRemoteConfig/url",
            "gitBranch": "*//scm/branches/hudson.plugins.git.BranchSpec/name",
            "pipelineFile": "*//scriptPath",
        }
    }

    def __init__(self) -> None:
        self._jenkins = jenkins.Jenkins(url=current_app.config["JENKINS_URL"],
                                        username=current_app.config["JENKINS_USER"],
                                        password=current_app.config["JENKINS_API_TOKEN"])

    def retrieve_job_parameters(self, build_id: str, build_number: int) -> list[Parameter]:
        job_info = self._jenkins.get_build_info(name=build_id, number=build_number)
        raw_config = self._jenkins.get_job_config(name=build_id)
        config = ET.fromstring(raw_config)
        parameter_defs = config.find("*//parameterDefinitions")
        if parameter_defs:
            descriptions = {
                define.findtext("name"): f"{define.findtext('description', '')}" + f" (default: <span class=\"fw-bold\">{define.findtext('defaultValue')}</span>)" if define.findtext('defaultValue') else ""
                for define in parameter_defs.iterfind("hudson.model.StringParameterDefinition")
            }
        else:
            descriptions = {}
        params = next(a for a in job_info["actions"] if a.get("_class", "#NONE") == "hudson.model.ParametersAction")["parameters"]
        params = [param for param in params if param["name"] != self.RESERVED_PARAMETER_NAME]
        for idx, param in enumerate(params):
            params[idx]["description"] = descriptions.get(param["name"], "")

        return params

    def get_releases_for_clone(self, test_id: str):
        test_id = UUID(test_id)
        # TODO: Filtering based on origin location / user preferences
        _: ArgusTest = ArgusTest.get(id=test_id)

        releases = list(ArgusRelease.all())

        return sorted(releases, key=lambda r: r.pretty_name if r.pretty_name else r.name)

    def get_groups_for_release(self, release_id: str):
        groups = list(ArgusGroup.filter(release_id=release_id).all())

        return sorted(groups, key=lambda g: g.pretty_name if g.pretty_name else g.name)

    def _verify_sct_settings(self, new_settings: dict[str, str]) -> tuple[bool, str]:
        if not (match := re.match(GITHUB_REPO_RE, new_settings["gitRepo"])):
            return (False, "Repository doesn't conform to GitHub schema")

        git_info = match.groupdict()
        if git_info.get("ssh"):
            repo = git_info["ssh_repo"]
            user = git_info["ssh_user"]
        else:
            repo = git_info["repo"]
            user = git_info["user"]

        user_tokens = UserOauthToken.filter(user_id=g.user.id).all()
        token = None
        for tok in user_tokens:
            if tok.kind == "github":
                token = tok.token
                break
        if not token:
            raise JenkinsServiceError("Github token not found")

        response = requests.get(
            url=f"https://api.github.com/repos/{user}/{repo}/contents/{new_settings['pipelineFile']}?ref={new_settings['gitBranch']}",
            headers={
                "Accept": "application/vnd.github+json",
                "Authorization": f"Bearer {token}",
            }
            )
        
        if response.status_code == 404:
            return (False, f"Pipeline file not found in the <a href=\"https://github.com/{user}/{repo}/tree/{new_settings['gitBranch']}\"> target repository</a>, please check the repository before continuing")

        if response.status_code == 403:
            return (True, "No access to this repository using your token. The pipeline file cannot be verified.")

        if response.status_code == 200:
            return (True, "")

        return (False, "Generic Error")

    def verify_job_settings(self, build_id: str, new_settings: dict[str, str]) -> tuple[bool, str]:
        PLUGIN_MAP = {
            "scylla-cluster-tests": self._verify_sct_settings,
            # for now they match
            "sirenada": self._verify_sct_settings,
            "driver-matrix-tests": self._verify_sct_settings,
        }
        test: ArgusTest = ArgusTest.get(build_system_id=build_id)
        plugin_name = test.plugin_name

        validated, message = PLUGIN_MAP.get(plugin_name, lambda _: (True, ""))(new_settings)

        return {
            "validated": validated,
            "message": message,
        }

    def get_advanced_settings(self, build_id: str):
        test: ArgusTest = ArgusTest.get(build_system_id=build_id)
        plugin_name = test.plugin_name

        if not (plugin_settings := self.SETTINGS_CONFIG_MAP.get(plugin_name)):
            return {}

        settings = {}
        raw_config = self._jenkins.get_job_config(name=build_id)
        config = ET.fromstring(raw_config)

        for setting, xpath in plugin_settings.items():
            value = config.find(xpath)
            settings[setting] = value.text

        return settings

    def adjust_job_settings(self, build_id: str, plugin_name: str, settings: dict[str, str]):
        xpath_map = self.SETTINGS_CONFIG_MAP.get(plugin_name)
        if not xpath_map:
            return
        
        config = self._jenkins.get_job_config(name=build_id)
        xml = ET.fromstring(config)
        for setting, value in settings.items():
            element = xml.find(xpath_map[setting])
            element.text = value

        adjusted_config = ET.tostring(xml, encoding="unicode")
        self._jenkins.reconfig_job(name=build_id, config_xml=adjusted_config)

    def clone_job(self, current_test_id: str, new_name: str, target: str, group: str, advanced_settings: bool | dict[str, str]):
        cloned_test: ArgusTest = ArgusTest.get(id=current_test_id)
        target_release: ArgusRelease = ArgusRelease.get(id=target)
        target_group: ArgusGroup = ArgusGroup.get(id=group)

        if target_group.id == cloned_test.id and new_name == cloned_test.name:
            raise JenkinsServiceError("Unable to clone: source and destination are the same")

        if not target_group.build_system_id:
            raise JenkinsServiceError("Unable to clone: target group is missing jenkins folder path")

        jenkins_new_build_id = f"{target_group.build_system_id}/{new_name}"

        new_test = ArgusTest()
        new_test.name = new_name
        new_test.build_system_id = jenkins_new_build_id
        new_test.group_id = target_group.id
        new_test.release_id = target_release.id
        new_test.plugin_name = cloned_test.plugin_name

        old_config = self._jenkins.get_job_config(name=cloned_test.build_system_id)
        LOGGER.info(old_config)
        xml = ET.fromstring(old_config)
        display_name = xml.find("displayName")
        if display_name:
            display_name.text = new_name
        new_config = ET.tostring(xml, encoding="unicode")
        self._jenkins.create_job(name=jenkins_new_build_id, config_xml=new_config)
        new_job_info = self._jenkins.get_job_info(name=jenkins_new_build_id)
        new_test.build_system_url = new_job_info["url"]
        new_test.save()

        if advanced_settings:
            self.adjust_job_settings(build_id=jenkins_new_build_id, plugin_name=new_test.plugin_name, settings=advanced_settings)

        return {
            "new_job": new_job_info,
            "new_entity": new_test,
        }

    def clone_build_job(self, build_id: str, params: dict[str, str]):
        queue_item = self.build_job(build_id=build_id, params=params)
        return {
            "queueItem": queue_item,
        }

    def build_job(self, build_id: str, params: dict, user_override: str = None):
        queue_number = self._jenkins.build_job(build_id, {
            **params,
            self.RESERVED_PARAMETER_NAME: g.user.username if not user_override else user_override
        })
        return queue_number

    def get_queue_info(self, queue_item: int):
        build_info = self._jenkins.get_queue_item(queue_item)
        LOGGER.info("%s", build_info)
        executable = build_info.get("executable")
        if executable:
            return executable
        else:
            return {
                "why": build_info["why"],
                "inQueueSince": build_info["inQueueSince"],
                "taskUrl": build_info["task"]["url"],
            }


import base64
from typing import Any
from uuid import UUID
from dataclasses import asdict
from argus.backend.plugins.sct.types import GeminiResultsRequest, PerformanceResultsRequest
from argus.backend.util.enums import ResourceState, TestStatus
from argus.client.base import ArgusClient
from argus.client.sct.types import EventsInfo, LogLink, Package


class ArgusSCTClient(ArgusClient):
    test_type = "scylla-cluster-tests"
    schema_version: None = "v8"

    class Routes(ArgusClient.Routes):
        SUBMIT_PACKAGES = "/sct/$id/packages/submit"
        SUBMIT_SCREENSHOTS = "/sct/$id/screenshots/submit"
        CREATE_RESOURCE = "/sct/$id/resource/create"
        TERMINATE_RESOURCE = "/sct/$id/resource/$name/terminate"
        UPDATE_RESOURCE = "/sct/$id/resource/$name/update"
        SET_SCT_RUNNER  = "/sct/$id/sct_runner/set"
        UPDATE_SHARDS_FOR_RESOURCE = "/sct/$id/resource/$name/shards"
        SUBMIT_NEMESIS = "/sct/$id/nemesis/submit"
        SUBMIT_GEMINI_RESULTS = "/sct/$id/gemini/submit"
        SUBMIT_PERFORMANCE_RESULTS = "/sct/$id/performance/submit"
        FINALIZE_NEMESIS = "/sct/$id/nemesis/finalize"
        SUBMIT_EVENTS = "/sct/$id/events/submit"
        SUBMIT_JUNIT_REPORT = "/sct/$id/junit/submit"

    def __init__(self, run_id: UUID, auth_token: str, base_url: str, api_version="v1") -> None:
        super().__init__(auth_token, base_url, api_version)
        self.run_id = run_id

    def submit_sct_run(self, job_name: str, job_url: str, started_by: str, commit_id: str,
                       origin_url: str, branch_name: str, sct_config: dict) -> None:
        """
            Submits an SCT run to argus.
        """
        response = super().submit_run(run_type=self.test_type, run_body={
            "run_id": str(self.run_id),
            "job_name": job_name,
            "job_url": job_url,
            "started_by": started_by,
            "commit_id": commit_id,
            "origin_url": origin_url,
            "branch_name": branch_name,
            "sct_config": sct_config,
        })

        self.check_response(response)

    def set_sct_run_status(self, new_status: TestStatus) -> None:
        """
            Sets an SCT run's status.
        """
        response = super().set_status(run_type=self.test_type, run_id=self.run_id, new_status=new_status)
        self.check_response(response)

    def set_sct_runner(self, public_ip: str, private_ip: str, region: str, backend: str) -> None:
        """
            Sets runner information for an SCT run.
        """
        response = self.post(
            endpoint=self.Routes.SET_SCT_RUNNER,
            location_params={"id": str(self.run_id)},
            body={
                **self.generic_body,
                "public_ip": public_ip,
                "private_ip": private_ip,
                "region": region,
                "backend": backend,
            }
        )
        self.check_response(response)

    def update_scylla_version(self, version: str) -> None:
        """
            Updates scylla server version used for filtering test results by version.
        """
        response = super().update_product_version(run_type=self.test_type, run_id=self.run_id, product_version=version)
        self.check_response(response)

    def submit_sct_logs(self, logs: list[LogLink]) -> None:
        """
            Submits links to logs collected from nodes by SCT
        """
        response = super().submit_logs(run_type=self.test_type, run_id=self.run_id, logs=logs)
        self.check_response(response)

    def finalize_sct_run(self) -> None:
        """
            Marks run as finished inside argus. Currently this only sets end time of the run,
            run status must be updated separately by .set_sct_run_status
        """
        response = super().finalize_run(run_type=self.test_type, run_id=self.run_id)
        self.check_response(response)

    def submit_packages(self, packages: list[Package]):
        """
            Submits packages, which are relevant to this test - kernel versions, scylla-server versions, etc
        """
        response = self.post(
            endpoint=self.Routes.SUBMIT_PACKAGES,
            location_params={"id": str(self.run_id)},
            body={
                **self.generic_body,
                "packages": [asdict(p) for p in packages],
            }
        )
        self.check_response(response)

    def submit_screenshots(self, screenshot_links: list[str]) -> None:
        """
            Submits links to the screenshots from grafana, taken at the end of the test.
        """
        response = self.post(
            endpoint=self.Routes.SUBMIT_SCREENSHOTS,
            location_params={"id": str(self.run_id)},
            body={
                **self.generic_body,
                "screenshot_links": screenshot_links,
            }
        )
        self.check_response(response)

    def submit_gemini_results(self, gemini_data: GeminiResultsRequest) -> None:
        """
            Submits gemini results such as oracle node information and gemini command & its results
        """
        response = self.post(
            endpoint=self.Routes.SUBMIT_GEMINI_RESULTS,
            location_params={"id": str(self.run_id)},
            body={
                **self.generic_body,
                "gemini_data": gemini_data,
            }
        )
        self.check_response(response)

    def submit_performance_results(self, performance_results: PerformanceResultsRequest) -> None:
        """
            Submits results of a performance run. Things such as throughput stats, overall and per op
        """
        response = self.post(
            endpoint=self.Routes.SUBMIT_PERFORMANCE_RESULTS,
            location_params={"id": str(self.run_id)},
            body={
                **self.generic_body,
                "performance_results": performance_results,
            }
        )
        self.check_response(response)

    def create_resource(self, name: str, resource_type: str, public_ip: str, private_ip: str,
                        region: str, provider: str, dc_name: str, rack_name: str, shards_amount: int, state=ResourceState.RUNNING) -> None:
        """
            Creates a cloud resource record in argus.
        """
        response = self.post(
            endpoint=self.Routes.CREATE_RESOURCE,
            location_params={"id": str(self.run_id)},
            body={
                **self.generic_body,
                "resource": {
                    "name": name,
                    "state": state,
                    "resource_type": resource_type,
                    "instance_details": {
                        "provider": provider,
                        "region": region,
                        "dc_name": dc_name,
                        "rack_name": rack_name,
                        "public_ip": public_ip,
                        "private_ip": private_ip,
                        "shards_amount": shards_amount,
                    }
                },
            }
        )
        self.check_response(response)

    def terminate_resource(self, name: str, reason: str) -> None:
        """
            Marks already created resource record as terminated.
        """
        response = self.post(
            endpoint=self.Routes.TERMINATE_RESOURCE,
            location_params={"id": str(self.run_id), "name": name},
            body={
                **self.generic_body,
                "reason": reason,
            }
        )
        self.check_response(response)

    def update_shards_for_resource(self, name: str, new_shards: int) -> None:
        """
            Separate methods to update "shards" of a resource, in cases
            where accurate reporting of shards is not possible during resource creation.
        """
        response = self.post(
            endpoint=self.Routes.UPDATE_SHARDS_FOR_RESOURCE,
            location_params={"id": str(self.run_id), "name": name},
            body={
                **self.generic_body,
                "shards": new_shards,
            }
        )
        self.check_response(response)


    def update_resource(self, name: str, update_data: dict[str, Any]) -> None:
        """
            Update fields of the resource.
        """
        response = self.post(
            endpoint=self.Routes.UPDATE_RESOURCE,
            location_params={"id": str(self.run_id), "name": name},
            body={
                **self.generic_body,
                "update_data": update_data,
            }
        )
        self.check_response(response)


    def submit_nemesis(self, name: str, class_name: str, start_time: int,
                       target_name: str, target_ip: str, target_shards: int) -> None:
        """
            Submits a nemesis record. Should then be finalized by
            .finalize_nemesis method on nemesis completion.
        """
        response = self.post(
            endpoint=self.Routes.SUBMIT_NEMESIS,
            location_params={"id": str(self.run_id)},
            body={
                **self.generic_body,
                "nemesis": {
                    "name": name,
                    "class_name": class_name,
                    "start_time": start_time,
                    "node_name": target_name,
                    "node_ip": target_ip,
                    "node_shards": target_shards,
                }
            }
        )
        self.check_response(response)

    def finalize_nemesis(self, name: str, start_time: int, status: str, message: str) -> None:
        """
            Finalizes an already submitted nemesis. Sets the status, nemesis message and end time.
        """
        response = self.post(
            endpoint=self.Routes.FINALIZE_NEMESIS,
            location_params={"id": str(self.run_id)},
            body={
                **self.generic_body,
                "nemesis": {
                    "name": name,
                    "start_time": start_time,
                    "status": status,
                    "message": message,
                }
            }
        )
        self.check_response(response)

    def submit_events(self, events: list[EventsInfo]) -> None:
        """
            Submits events collected by event collector once at the end of the test.
        """
        response = self.post(
            endpoint=self.Routes.SUBMIT_EVENTS,
            location_params={"id": str(self.run_id)},
            body={
                **self.generic_body,
                "events": [asdict(e) for e in events]
            }
        )
        self.check_response(response)

    def sct_heartbeat(self) -> None:
        """
            Updates a heartbeat field for an already existing test.
        """
        super().heartbeat(run_type=self.test_type, run_id=self.run_id)

    def sct_submit_junit_report(self, file_name: str, raw_content: str) -> None:
        """
            Submits a JUnit-formatted XML report to argus
        """
        response = self.post(
            endpoint=self.Routes.SUBMIT_JUNIT_REPORT,
            location_params={"id": str(self.run_id)},
            body={
                **self.generic_body,
                "file_name": file_name,
                "content": str(base64.encodebytes(bytes(raw_content, encoding="utf-8")), encoding="utf-8")
            }
        )


import datetime
import logging
import re
from functools import partial, reduce
from typing import Any, Callable, TypedDict
from urllib.parse import unquote
from uuid import UUID

from cassandra.cqlengine.models import Model
from argus.backend.models.web import ArgusGroup, ArgusRelease, ArgusTest, ArgusUserView, User
from argus.backend.plugins.loader import all_plugin_models
from argus.backend.util.common import chunk, current_user

LOGGER = logging.getLogger(__name__)


class UserViewException(Exception):
    pass


class ViewUpdateRequest(TypedDict):
    name: str
    description: str
    display_name: str
    tests: list[str]
    widget_settings: str


class UserViewService:
    ADD_ALL_ID = UUID("db6f33b2-660b-4639-ba7f-79725ef96616")
    def create_view(self, name: str, items: list[str], widget_settings: str, description: str = None, display_name: str = None) -> ArgusUserView:
        try:
            name_check = ArgusUserView.get(name=name)
            raise UserViewException(f"View with name {name} already exists: {name_check.id}", name, name_check, name_check.id)
        except ArgusUserView.DoesNotExist:
            pass
        view = ArgusUserView()
        view.name = name
        view.display_name = display_name or name
        view.description = description
        view.widget_settings = widget_settings
        view.tests = []
        for entity in items:
            entity_type, entity_id = entity.split(":")
            match (entity_type):
                case "release": 
                    view.tests.extend(t.id for t in ArgusTest.filter(release_id=entity_id).all())
                    view.release_ids.append(entity_id)
                case "group": 
                    view.tests.extend(t.id for t in ArgusTest.filter(group_id=entity_id).all())
                    view.group_ids.append(entity_id)
                case "test": 
                    view.tests.append(entity_id)
        view.user_id = current_user().id

        view.save()
        return view

    @staticmethod
    def index_mapper(item: Model, type = "test"):
        mapped = dict(item)
        mapped["type"] = type
        return mapped

    def test_lookup(self, query: str):
        def check_visibility(entity: dict):
            if not entity["enabled"]:
                return False
            if entity.get("group") and not entity["group"]["enabled"]:
                return False
            if entity.get("release") and not entity["release"]["enabled"]:
                return False
            return True

        def facet_extraction(query: str) -> str:
            extractor = re.compile(r"(?:(?P<name>(?:release|group|type)):(?P<value>\"?[\w\d\.\-]*\"?))")
            facets = re.findall(extractor, query)

            return (re.sub(extractor, "", query).strip(), facets)

        def type_facet_filter(item: dict, key: str, facet_query: str):
            entity_type: str = item[key]
            return facet_query.lower() == entity_type

        def facet_filter(item: dict, key: str, facet_query: str):
            if entity := item.get(key):
                name: str = entity.get("pretty_name") or entity.get("name")
                return facet_query.lower() in name.lower() if name else False
            return False

        def facet_wrapper(query_func: Callable[[dict], bool], facet_query: str, facet_type: str) -> bool:
            def inner(item: dict, query: str):
                return query_func(item, query) and facet_funcs[facet_type](item, facet_type, facet_query)
            return inner

        facet_funcs = {
            "type": type_facet_filter,
            "release": facet_filter,
            "group": facet_filter,
        }

        def index_searcher(item, query: str):
            name: str = item["pretty_name"] or item["name"]
            return unquote(query).lower() in name.lower() if query else True

        text_query, facets = facet_extraction(query)
        search_func = index_searcher
        for facet, value in facets:
            if facet in facet_funcs.keys():
                search_func = facet_wrapper(query_func=search_func, facet_query=value, facet_type=facet)


        all_tests = ArgusTest.all()
        all_releases = ArgusRelease.all()
        all_groups = ArgusGroup.all()
        release_by_id = {release.id: partial(self.index_mapper, type="release")(release) for release in all_releases}
        group_by_id = {group.id: partial(self.index_mapper, type="group")(group) for group in all_groups}
        index = [self.index_mapper(t) for t in all_tests]
        index = [*release_by_id.values(), *group_by_id.values(), *index]
        for item in index:
            item["group"] = group_by_id.get(item.get("group_id"))
            item["release"] = release_by_id.get(item.get("release_id"))

        results = filter(partial(search_func, query=text_query), index)

        return [{ "id": self.ADD_ALL_ID, "name": "Add all...", "type": "special" }, *list(res for res in results if check_visibility(res))]

    def update_view(self, view_id: str | UUID, update_data: ViewUpdateRequest) -> bool:
        view: ArgusUserView = ArgusUserView.get(id=view_id)
        if view.user_id != current_user().id and not current_user().is_admin():
            raise UserViewException("Unable to modify other users' views")
        for key in ["user_id", "id"]:
            update_data.pop(key, None)
        items = update_data.pop("items")
        for k, value in update_data.items():
            view[k] = value
        view.tests = []
        view.release_ids = []
        view.group_ids = []
        for entity in items:
            entity_type, entity_id = entity.split(":")
            match (entity_type):
                case "release": 
                    view.tests.extend(t.id for t in ArgusTest.filter(release_id=entity_id).all())
                    view.release_ids.append(entity_id)
                case "group": 
                    view.tests.extend(t.id for t in ArgusTest.filter(group_id=entity_id).all())
                    view.group_ids.append(entity_id)
                case "test": 
                    view.tests.append(entity_id)
        view.last_updated = datetime.datetime.utcnow()
        view.save()
        return True

    def delete_view(self, view_id: str | UUID) -> bool:
        view = ArgusUserView.get(id=view_id)
        if view.user_id != current_user().id and not current_user().is_admin():
            raise UserViewException("Unable to modify other users' views")
        view.delete()

        return True

    def get_view(self, view_id: str | UUID) -> ArgusUserView:
        view: ArgusUserView = ArgusUserView.get(id=view_id)
        if datetime.datetime.utcnow() - (view.last_updated or datetime.datetime.fromtimestamp(0)) > datetime.timedelta(hours=1):
            self.refresh_stale_view(view)
        return view

    def get_view_by_name(self, view_name: str) -> ArgusUserView:
        view: ArgusUserView = ArgusUserView.get(name=view_name)
        if datetime.datetime.utcnow() - (view.last_updated or datetime.datetime.fromtimestamp(0)) > datetime.timedelta(hours=1):
            self.refresh_stale_view(view)
        return view

    def get_all_views(self, user: User | None = None) -> list[ArgusUserView]:
        if user:
            return list(ArgusUserView.filter(user_id=user.id).all())
        return list(ArgusUserView.filter().all())

    def resolve_view_tests(self, view_id: str | UUID) -> list[ArgusTest]:
        view = ArgusUserView.get(id=view_id)
        return self.resolve_tests_by_id(view.tests)

    def resolve_tests_by_id(self, test_ids: list[str | UUID]) -> list[ArgusTest]:
        tests = []
        for batch in chunk(test_ids):
            tests.extend(ArgusTest.filter(id__in=batch).all())

        return tests

    def batch_resolve_entity(self, entity: Model, param_name: str, entity_ids: list[UUID]) -> list[Model]:
        result = []
        for batch in chunk(entity_ids):
            result.extend(entity.filter(**{f"{param_name}__in": batch}).allow_filtering().all())
        return result

    def refresh_stale_view(self, view: ArgusUserView):
        view.tests = [test.id for test in self.resolve_view_tests(view.id)]
        all_tests = set(view.tests)
        all_tests.update(test.id for test in self.batch_resolve_entity(ArgusTest, "group_id", view.group_ids))
        all_tests.update(test.id for test in self.batch_resolve_entity(ArgusTest, "release_id", view.release_ids))
        view.tests = list(all_tests)
        view.last_updated = datetime.datetime.utcnow()
        view.save()

        return view

    def resolve_releases_for_tests(self, tests: list[ArgusTest]):
        releases = []
        unique_release_ids = reduce(lambda releases, test: releases.add(test.release_id) or releases, tests, set())
        for batch in chunk(unique_release_ids):
            releases.extend(ArgusRelease.filter(id__in=batch).all())

        return releases

    def resolve_groups_for_tests(self, tests: list[ArgusTest]):
        releases = []
        unique_release_ids = reduce(lambda groups, test: groups.add(test.group_id) or groups, tests, set())
        for batch in chunk(unique_release_ids):
            releases.extend(ArgusGroup.filter(id__in=batch).all())

        return releases

    def get_versions_for_view(self, view_id: str | UUID) -> list[str]:
        tests = self.resolve_view_tests(view_id)
        unique_versions = {ver for plugin in all_plugin_models()
                           for ver in plugin.get_distinct_versions_for_view(tests=tests)}

        return sorted(list(unique_versions), reverse=True)

    def resolve_view_for_edit(self, view_id: str | UUID) -> dict:
        view: ArgusUserView = ArgusUserView.get(id=view_id)
        resolved = dict(view)
        view_groups = self.batch_resolve_entity(ArgusGroup, "id", view.group_ids)
        view_releases = self.batch_resolve_entity(ArgusRelease, "id", view.release_ids)
        view_tests = self.resolve_view_tests(view.id)
        all_groups = { group.id: partial(self.index_mapper, type="group")(group) for group in self.resolve_releases_for_tests(view_tests) }
        all_releases ={ release.id: partial(self.index_mapper, type="release")(release) for release in self.resolve_releases_for_tests(view_tests) }
        entities_by_id = {
            entity.id: partial(self.index_mapper, type="release" if isinstance(entity, ArgusRelease) else "group")(entity)
            for container in [view_releases, view_groups]
            for entity in container
        }

        items = []
        for test in view_tests:
            if not (entities_by_id.get(test.group_id) or entities_by_id.get(test.release_id)):
                item = dict(test)
                item["type"] = "test"
                items.append(item)

        items = [*entities_by_id.values(), *items]
        for entity in items:
            entity["group"] = all_groups.get(entity.get("group_id"), {}).get("pretty_name") or all_groups.get(entity.get("group_id"), {}).get("name")
            entity["release"] = all_releases.get(entity.get("release_id"), {}).get("pretty_name") or all_releases.get(entity.get("release_id"), {}).get("name")

        resolved["items"] = items
        return resolved



from dataclasses import dataclass, field, asdict
from typing import Dict, List, Tuple, Union, Type
from uuid import UUID
from enum import Enum
import json



class Status(Enum):
    PASS = 1
    WARNING = 2
    ERROR = 3


class ResultType(Enum):
    INTEGER = int
    FLOAT = float
    TEXT = str


@dataclass
class ColumnMetadata:
    id: int
    unit: str
    type: ResultType


@dataclass
class Result:
    value: Union[int, float, str]
    status: Status


class ResultTableMeta(type):
    def __new__(cls, name, bases, dct):
        cls_instance = super().__new__(cls, name, bases, dct)
        meta = dct.get('Meta')

        if meta:
            cls_instance.table_name = meta.table_name
            cls_instance.columns_map = {col_name: ColumnMetadata(id=col_id, unit=unit, type=result_type)
                                        for col_name, (col_id, unit, result_type) in meta.Columns.items()}
            cls_instance.rows_map = {row_name: row_id
                                     for row_name, row_id in meta.Rows.items()}
        return cls_instance


class Row:
    def __init__(self, columns_map: Dict[str, ColumnMetadata]):
        self.columns_map = columns_map
        self.data: Dict[str, Result] = {}

    def __getitem__(self, column_name: str) -> Result:
        return self.data[column_name]

    def __setitem__(self, column_name: str, result: Result):
        if column_name not in self.columns_map:
            raise ValueError(f"Column name '{column_name}' not found in columns_map.")
        column_metadata = self.columns_map[column_name]
        if not isinstance(result.value, column_metadata.type.value):
            raise ValueError(f"Value {result.value} for column '{column_name}' is not of type {column_metadata.type.name}")
        self.data[column_name] = result


@dataclass
class BaseResultTable(metaclass=ResultTableMeta):
    results: Dict[str, Row] = field(default_factory=dict)

    def as_dict(self) -> dict:
        results_list = []
        for row_name, row in self.results.items():
            row_id = self.rows_map.get(row_name)
            for column_name, result in row.data.items():
                column_metadata = self.columns_map.get(column_name)
                results_list.append({
                    "column_id": column_metadata.id,
                    "row_id": row_id,
                    "result": result.value,
                    "status": result.status.value
                })

        meta_info = {
            "table_name": self.table_name,
            "columns": {name: {"id": meta.id, "unit": meta.unit, "type": meta.type.name} for name, meta in self.columns_map.items()},
            "rows": self.rows_map
        }
        return {
            "meta": meta_info,
            "results": results_list
        }

    def to_json(self) -> str:
        dict_result = self.as_dict()
        return json.dumps(dict_result, default=str)

    def __getitem__(self, row_name: str) -> Row:
        if row_name not in self.rows_map:
            raise ValueError(f"Row name '{row_name}' not found in rows_map.")
        if row_name not in self.results:
            self.results[row_name] = Row(self.columns_map)
        return self.results[row_name]


# Example of a specific result table with its own metadata
class LatencyResultTable(BaseResultTable):
    class Meta:
        table_name = "latency_percentile_write"
        Columns = {
            "latency": (1, "ms", ResultType.FLOAT),
            "op_rate": (2, "ops", ResultType.INTEGER)
        }
        Rows = {
            "mean": 1,
            "p99": 2
        }


# Client code example
if __name__ == "__main__":
    class LatencyResultTable(BaseResultTable):
        class Meta:
            table_name = "latency_percentile_write"
            Columns = {
                "latency": (1, "ms", ResultType.FLOAT),
                "op_rate": (2, "ops", ResultType.INTEGER)
            }
            Rows = {
                "mean": 1,
                "p99": 2
            }

    result_table = LatencyResultTable()

    result_table["mean"]["latency"] = Result(value=1.1, status=Status.WARNING)
    result_table["mean"]["op_rate"] = Result(value=59988, status=Status.ERROR)
    result_table["p99"]["latency"] = Result(value=2.7, status=Status.PASS)
    result_table["p99"]["op_rate"] = Result(value=59988, status=Status.WARNING)

    from argus.client.sct.client import ArgusSCTClient

    run_id = UUID("24e09748-bba4-47fd-a615-bf7ea2c425eb")
    client = ArgusSCTClient(run_id, auth_token="UO+2GXL9XqSgcVJijWk5WnbPXPit5ot5nfkLAHAr7SaqROfSCWycabpp/wxyY8+I", base_url="http://localhost:5000")
    client.submit_results(result_table)


import logging
from uuid import UUID
from flask import (
    Blueprint,
    request
)

from argus.backend.error_handlers import handle_api_exception
from argus.backend.models.web import ArgusTest
from argus.backend.service.jenkins_service import JenkinsService
from argus.backend.service.testrun import TestRunService
from argus.backend.service.user import api_login_required
from argus.backend.util.common import get_payload
from argus.backend.util.enums import TestInvestigationStatus, TestStatus

bp = Blueprint('testrun_api', __name__, 'testrun')
bp.register_error_handler(Exception, handle_api_exception)
LOGGER = logging.getLogger(__name__)


@bp.route("/test/<string:test_id>/runs")
@api_login_required
def get_runs_for_test(test_id: str):
    limit = request.args.get("limit")
    if not limit:
        limit = 10
    else:
        limit = int(limit)
    additional_runs = [UUID(run) for run in request.args.getlist('additionalRuns[]')]
    service = TestRunService()

    runs = service.get_runs_by_test_id(test_id=UUID(test_id), additional_runs=additional_runs, limit=limit)

    return {
        "status": "ok",
        "response": runs
    }


@bp.route("/run/<string:run_type>/<string:run_id>")
@api_login_required
def get_testrun(run_type: str, run_id: str):
    run_id = UUID(run_id)
    service = TestRunService()
    test_run = service.get_run(run_type=run_type, run_id=run_id)

    return {
        "status": "ok",
        "response": test_run
    }


@bp.route("/run/<string:run_id>/activity")
@api_login_required
def test_run_activity(run_id: str):
    run_id = UUID(run_id)
    service = TestRunService()
    activity = service.get_run_events(run_id=run_id)

    return {
        "status": "ok",
        "response": activity
    }



@bp.route("/run/<string:test_id>/<string:run_id>/fetch_results", methods=["GET"])
@api_login_required
def fetch_results(test_id: str, run_id: str):
    tables = TestRunService().fetch_results(test_id=UUID(test_id), run_id=UUID(run_id))
    return {
        "status": "ok",
        "tables": tables
    }


@bp.route("/test/<string:test_id>/run/<string:run_id>/status/set", methods=["POST"])
@api_login_required
def set_testrun_status(test_id: str, run_id: str):
    payload = get_payload(client_request=request)
    new_status = payload.get("status")
    if not new_status:
        raise Exception("Status not specified in the request")
    service = TestRunService()
    result = service.change_run_status(
        test_id=UUID(test_id),
        run_id=UUID(run_id),
        new_status=TestStatus(new_status)
    )

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/test/<string:test_id>/run/<string:run_id>/investigation_status/set", methods=["POST"])
@api_login_required
def set_testrun_investigation_status(test_id: str, run_id: str):
    payload = get_payload(client_request=request)
    new_status = payload.get("investigation_status")
    if not new_status:
        raise Exception("Status not specified in the request")
    service = TestRunService()
    result = service.change_run_investigation_status(
        test_id=UUID(test_id),
        run_id=UUID(run_id),
        new_status=TestInvestigationStatus(new_status),
    )

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/test/<string:test_id>/run/<string:run_id>/assignee/set", methods=["POST"])
@api_login_required
def set_testrun_assignee(test_id: str, run_id: str):
    payload = get_payload(client_request=request)
    assignee = payload.get("assignee")
    if not assignee:
        raise Exception("Assignee not specified in the request")
    service = TestRunService()
    result = service.change_run_assignee(
        test_id=UUID(test_id),
        run_id=UUID(run_id),
        new_assignee=UUID(assignee) if assignee != TestRunService.ASSIGNEE_PLACEHOLDER else None,
    )

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/test/<string:test_id>/run/<string:run_id>/issues/submit", methods=["POST"])
@api_login_required
def issues_submit(test_id: str, run_id: str):
    payload = get_payload(client_request=request)
    service = TestRunService()
    submit_result = service.submit_github_issue(
        issue_url=payload["issue_url"],
        test_id=UUID(test_id),
        run_id=UUID(run_id)
    )

    return {
        "status": "ok",
        "response": submit_result
    }


@bp.route("/issues/get", methods=["GET"])
@api_login_required
def issues_get():
    filter_key = request.args.get("filterKey")
    if not filter_key:
        raise Exception("Filter key not provided in the request")
    key_value = request.args.get("id")
    if not key_value:
        raise Exception("Id wasn't provided in the request")
    key_value = UUID(key_value)
    aggregate_by_issue = request.args.get("aggregateByIssue")
    aggregate_by_issue = bool(int(aggregate_by_issue)) if aggregate_by_issue else False
    service = TestRunService()
    issues = service.get_github_issues(
        filter_key=filter_key,
        filter_id=key_value,
        aggregate_by_issue=aggregate_by_issue
    )

    return {
        "status": "ok",
        "response": issues
    }


@bp.route("/issues/delete", methods=["POST"])
@api_login_required
def issues_delete():
    payload = get_payload(client_request=request)
    service = TestRunService()
    result = service.delete_github_issue(issue_id=payload["issue_id"])

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/run/<string:run_id>/comments", methods=["GET"])
@api_login_required
def get_testrun_comments(run_id: str):
    service = TestRunService()
    comments = service.get_run_comments(
        run_id=UUID(run_id)
    )

    return {
        "status": "ok",
        "response": comments
    }


@bp.route("/comment/<string:comment_id>/get", methods=["GET"])
@api_login_required
def get_single_comment(comment_id: str):
    service = TestRunService()
    comment = service.get_run_comment(
        comment_id=UUID(comment_id)
    )

    return {
        "status": "ok",
        "response": comment
    }


@bp.route("/test/<string:test_id>/run/<string:run_id>/comments/submit", methods=["POST"])
@api_login_required
def submit_testrun_comment(test_id: str, run_id: str):
    payload = get_payload(request)
    service = TestRunService()
    result = service.post_run_comment(
        test_id=UUID(test_id),
        run_id=UUID(run_id),
        message=payload["message"],
        reactions=payload["reactions"],
        mentions=payload["mentions"],
    )

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/test/<string:test_id>/run/<string:run_id>/comment/<string:comment_id>/update", methods=["POST"])
@api_login_required
def test_run_update_comment(test_id: str, run_id: str, comment_id: str):
    payload = get_payload(request)
    service = TestRunService()
    result = service.update_run_comment(
        test_id=UUID(test_id),
        run_id=UUID(run_id),
        comment_id=UUID(comment_id),
        message=payload["message"],
        reactions=payload["reactions"],
        mentions=payload["mentions"],
    )

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/test/<string:test_id>/run/<string:run_id>/comment/<string:comment_id>/delete", methods=["POST"])
@api_login_required
def test_run_delete_comment(test_id: str, run_id: str, comment_id: str):
    service = TestRunService()
    result = service.delete_run_comment(
        test_id=UUID(test_id),
        run_id=UUID(run_id),
        comment_id=UUID(comment_id),
    )

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/terminate_stuck_runs", methods=["POST"])
@api_login_required
def sct_terminate_stuck_runs():
    result = TestRunService().terminate_stuck_runs()
    return {
        "status": "ok",
        "response": {
            "total": result
        }
    }


@bp.route("/ignore_jobs", methods=["POST"])
@api_login_required
def ignore_jobs():
    payload = get_payload(request)
    service = TestRunService()

    result = service.ignore_jobs(test_id=payload["testId"], reason=payload["reason"])

    return {
        "status": "ok",
        "response": {
            "affectedJobs": result
        }
    }


@bp.route("/get_runs_by_test_id_run_id", methods=["POST"])
@api_login_required
def get_runs_by_test_id_run_id():
    payload: list[tuple[UUID, UUID]] = get_payload(request)
    service = TestRunService()

    result = service.resolve_run_build_id_and_number_multiple(payload)

    return {
        "status": "ok",
        "response": {
            "runs": result
        }
    }


@bp.route("/jenkins/params", methods=["POST"])
@api_login_required
def get_jenkins_job_params():

    payload = get_payload(request)
    service = JenkinsService()

    result = service.retrieve_job_parameters(build_id=payload["buildId"], build_number=payload["buildNumber"])

    return {
        "status": "ok",
        "response": {
            "parameters": result
        }
    }


@bp.route("/jenkins/build", methods=["POST"])
@api_login_required
def build_jenkins_job():

    payload = get_payload(request)
    service = JenkinsService()

    result = service.build_job(build_id=payload["buildId"], params=payload["parameters"])

    return {
        "status": "ok",
        "response": {
            "queueItem": result
        }
    }


@bp.route("/jenkins/queue_info")
@api_login_required
def get_queue_info():
    queue_item = request.args.get("queueItem")
    if not queue_item:
        raise Exception("No queueItem provided")
    service = JenkinsService()
    result = service.get_queue_info(int(queue_item))

    return {
        "status": "ok",
        "response": {
            "queueItem": result
        }
    }


@bp.route("/jenkins/clone/targets")
@api_login_required
def get_clone_targets():
    test_id = request.args.get("testId")
    if not test_id:
        raise Exception("No testId provided")
    service = JenkinsService()
    result = service.get_releases_for_clone(test_id)

    return {
        "status": "ok",
        "response": {
            "targets": result
        }
    }


@bp.route("/jenkins/clone/groups")
@api_login_required
def get_groups_for_target():
    target_id = request.args.get("targetId")
    if not target_id:
        raise Exception("No targetId provided")
    service = JenkinsService()
    result = service.get_groups_for_release(target_id)

    return {
        "status": "ok",
        "response": {
            "groups": result
        }
    }


@bp.route("/jenkins/clone/create", methods=["POST"])
@api_login_required
def clone_jenkins_job():

    payload = get_payload(request)
    service = JenkinsService()

    result = service.clone_job(
        current_test_id=payload["currentTestId"],
        new_name=payload["newName"],
        target=payload["target"],
        group=payload["group"],
        advanced_settings=payload["advancedSettings"],
    )

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/jenkins/clone/build", methods=["POST"])
@api_login_required
def clone_build_jenkins_job():

    payload = get_payload(request)
    service = JenkinsService()

    result = service.clone_build_job(build_id=payload["buildId"], params=payload["parameters"])

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/jenkins/clone/settings")
@api_login_required
def get_clone_job_advanced_settings():
    build_id = request.args.get("buildId")
    if not build_id:
        raise Exception("No testId provided")
    service = JenkinsService()
    result = service.get_advanced_settings(build_id)

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/jenkins/clone/settings/change", methods=["POST"])
@api_login_required
def set_job_settings():
    payload = get_payload(request)
    service = JenkinsService()
    test = ArgusTest.get(build_system_id=payload["buildId"])
    result = service.adjust_job_settings(build_id=test.build_system_id, plugin_name=test.plugin_name, settings=payload["settings"])

    return {
        "status": "ok",
        "response": result
    }


@bp.route("/jenkins/clone/settings/validate", methods=["POST"])
@api_login_required
def clone_validate_new_settings():

    payload = get_payload(request)
    service = JenkinsService()

    result = service.verify_job_settings(build_id=payload["buildId"], new_settings=payload["newSettings"])

    return {
        "status": "ok",
        "response": result
    }



from cassandra.cqlengine import columns
from cassandra.cqlengine.models import Model
from cassandra.cqlengine.usertype import UserType
from enum import Enum


class Status(Enum):
    PASS = 0
    WARNING = 1
    ERROR = 2


class ColumnMetadata(UserType):
    name = columns.Ascii()
    unit = columns.Text()
    type = columns.Ascii()


class ArgusGenericResultMetadata(Model):
    __table_name__ = "generic_result_metadata_v1"
    test_id = columns.UUID(partition_key=True)
    name = columns.Text(required=True, primary_key=True)
    description = columns.Text()
    columns_meta = columns.List(value_type=columns.UserDefinedType(ColumnMetadata))
    rows_meta = columns.List(value_type=columns.Ascii())

    def __init__(self, **kwargs):
        kwargs["columns_meta"] = [ColumnMetadata(**col) for col in kwargs.pop('columns_meta', [])]
        super().__init__(**kwargs)


class ArgusGenericResultData(Model):
    __table_name__ = "generic_result_data_v1"
    test_id = columns.UUID(partition_key=True)
    name = columns.Text(partition_key=True)
    run_id = columns.UUID(primary_key=True)
    column = columns.Ascii(primary_key=True, index=True)
    row = columns.Ascii(primary_key=True, index=True)
    sut_timestamp = columns.DateTime()  # for sorting
    value = columns.Double()
    status = columns.Ascii()
